import copy
import json
import os
import time

from remotemanager.storage.database import Database
from remotemanager.storage import SendableMixin, TrackedFile
from remotemanager.storage.sendablemixin import SERIALISED_STORAGE_KEY
from remotemanager.logging.utils import format_iterable
from remotemanager.utils.uuid import generate_uuid
from remotemanager.utils import object_from_uuid, _time_format, ensure_list
from remotemanager.logging import LoggingMixin
from remotemanager.logging.verbosity import Verbosity

from datetime import datetime

localwinerror = """Local runs on windows machines are not supported.
Please use a URL which connects to a non-windows machine or consider using
Docker to continue."""


class Runner(SendableMixin, LoggingMixin):
    """
    The Runner class stores any info pertaining to this specific run. E.g.
    Arguments, result, run status, files, etc.

    .. warning::
        Interacting with this object directly could cause unstable behaviour.
        It is best to allow Dataset to handle the runners. If you require a
        single run, you should create a Dataset and append just that one run.
    """

    _defaults = {"skip": True}

    _default_local_dir = "temp_runner_local"
    _default_remote_dir = "temp_runner_remote"

    _submit_status = {
        "initial": "creating files and submitting...",
        "final": "command executed remotely",
    }

    _args_replaced_key = "~serialised_args~"

    _do_not_package = ["_serialiser", "_parent", "_database"]

    def __init__(
        self,
        arguments: dict,
        dbfile: str,
        parent,
        self_id: str,
        extra_files_send: list = None,
        extra_files_recv: list = None,
        verbose: int = None,
        **kwargs,
    ):

        self._verbose = Verbosity(verbose)

        self._run_options = self._set_defaults(kwargs)

        self._extra_files = {
            "send": extra_files_send if extra_files_send is not None else [],
            "recv": extra_files_recv if extra_files_recv is not None else [],
        }

        if arguments is not None and not isinstance(arguments, dict):
            raise ValueError(f"runner arguments ({type(arguments)}) must be dict-type")

        # parent and id setting
        self._parent = parent
        self._parent_uuid = parent.uuid  # used for parent memory recovery

        self._id = self_id

        # check that we can properly serialise the args
        # this needs to be within the runner, so we can properly generate uuids
        self._args_replaced = False
        try:
            json.dumps(arguments)
            uuid_slug = copy.deepcopy(arguments) or {}
        except TypeError:
            file = f"{self.parent.argfile}-{self.id}{self.serialiser.extension}"
            lpath = os.path.join(self.parent.local_dir, file)

            if not os.path.isdir(self.parent.local_dir):
                os.makedirs(self.parent.local_dir)

            content = self.parent.serialiser.dumps(arguments)
            with open(lpath, self.serialiser.write_mode) as o:
                o.write(content)

            # adding the file in here forces the kwargs to swap out
            # kwargs for a repo.load
            arguments = {file: Runner._args_replaced_key}
            self._args_replaced = True

            uuid_slug = {"uuid_base": generate_uuid(str(content))}

            extra_files_send.append(lpath)

        # uuid generation
        uuid_slug.update(**kwargs)
        self._runner_uuid = generate_uuid(format_iterable(uuid_slug))
        self._uuid = generate_uuid(self._runner_uuid + str(self.parent.uuid))

        self._logger.info(f"new runner (id {self.uuid}) created")

        self._args = arguments
        self._state_time = None
        self._last_submitted = -1
        self._state = None
        self._extension = "yaml"

        self._dependency_info = {}

        self._dbfile = dbfile

        self._manual_run = False
        self._history = {}
        self.state = "created"

        self._identifier = f"{self.parent.name}-{self.parent.short_uuid}-{self.id}"

    def __hash__(self) -> hash:
        return hash(self.uuid)

    def __repr__(self) -> str:
        return self.identifier

    @property
    def verbose(self) -> Verbosity:
        return self._verbose

    @property
    def database(self) -> Database:
        """
        Access to the stored database object.
        Creates a connection if none exist.

        Returns:
            Database
        """
        if not hasattr(self, "_database"):
            self._database = Database(file=self._dbfile)
        return self._database

    @property
    def parent(self):
        if self.is_missing("_parent"):
            self._parent = object_from_uuid(self._parent_uuid, "Dataset")
        return self._parent

    @property
    def serialiser(self):
        return self.parent.serialiser

    @staticmethod
    def _set_defaults(kwargs: dict = None) -> dict:
        """
        Sets default arguments as expected. If used as a staticmethod, returns
        the defaults
        """

        if kwargs is None:
            kwargs = {}

        for k, v in Runner._defaults.items():
            if k not in kwargs:
                kwargs[k] = v

        return kwargs

    @property
    def uuid(self) -> str:
        """
        The uuid of this runner
        """
        return self._uuid

    @property
    def short_uuid(self) -> str:
        """
        A short uuid for filenames
        """
        return self.uuid[:8]

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._id

    @property
    def identifier(self) -> str:
        return self._identifier

    def _format_filename(self, ftype: str, ext: str) -> str:
        """
        Formats internal file names consistently.

        Args:
            ftype (str):
                file type. Jobscript, result file, etc.
            ext (str):
                file extension

        Returns:
            str: formatted filename
        """
        return f"{self.identifier}-{ftype}{ext}"

    @property
    def runfile(self) -> TrackedFile:
        """
        Filename of the python runfile
        """
        base_name = self._format_filename("run", ".py")

        return TrackedFile(self.local_dir, self.remote_dir, base_name)

    @property
    def jobscript(self) -> TrackedFile:
        """
        Filename of the run script
        """
        base_name = self._format_filename("jobscript", ".sh")

        return TrackedFile(self.local_dir, self.remote_dir, base_name)

    @property
    def resultfile(self) -> TrackedFile:
        """
        Result file name
        """
        base_name = self._format_filename("result", f".{self.result_extension}")

        paths = [self.remote_dir]
        if self.run_dir != self.remote_dir and self.run_dir is not None:
            paths.append(self.run_dir)
        joined = "/".join(paths)

        return TrackedFile(self.local_dir, joined, base_name)

    @property
    def errorfile(self) -> TrackedFile:
        """
        File tracker for error dumpfile
        """
        base_name = self._format_filename("error", ".out")

        return TrackedFile(self.local_dir, self.remote_dir, base_name)

    def _relative_remote_dir(self, file: str) -> str:
        """
        remote dir path relative to run dir
        """
        self._logger.info(
            f"getting relative path to remote dir for file "
            f"{file} where remote={self.remote_dir} and run="
            f"{self.run_dir}"
        )
        if self.run_dir is None:
            self._logger.info("no run dir, returning file as rel path")
            return file
        return os.path.join("..", file)

    @property
    def result_extension(self) -> str:
        """
        Resultfile file format extension
        """
        return self._extension

    @result_extension.setter
    def result_extension(self, ext: str) -> None:
        """
        Sets the resultfile format extension

        .. warning::
            This does not change anything aside from the extension that
            the runner looks for when trying to find a result file. If you
            require a different serialisation, you should set the serialiser.
        """
        self._extension = ext.strip(".")

    @property
    def local_dir(self) -> str:
        """
        Local staging directory
        """
        return self._run_options.get("local_dir", Runner._default_local_dir)

    @local_dir.setter
    def local_dir(self, path: str) -> None:
        """
        Sets the local_dir
        """
        self._run_options["local_dir"] = path

    @property
    def remote_dir(self) -> str:
        """
        Target directory on the remote for transports
        """
        if "remote_dir" in self._run_options:
            return self._run_options["remote_dir"]
        return self._run_options.get("run_dir", Runner._default_remote_dir)

    @remote_dir.setter
    def remote_dir(self, path: str) -> None:
        """
        Sets the remote_dir
        """
        self._logger.debug(f"setting remote dir to {path}")
        self._run_options["remote_dir"] = path

    @property
    def run_dir(self) -> [str, None]:
        """
        Intended running directory. If not set, uses remote_dir

        .. note::
            If both remote_dir and run_dir are set, the files will be
            transferred to remote_dir, and then executed within run_dir
        """
        if "run_dir" in self._run_options:
            abs_rundir = self._run_options["run_dir"]
            common = os.path.commonpath((self.remote_dir, abs_rundir))

            run = os.path.split(abs_rundir.replace(common, ""))[1]
            return run
        return None

    @run_dir.setter
    def run_dir(self, dir: str) -> None:
        """
        Sets the run_dir
        """
        self._run_options["run_dir"] = dir

    @property
    def args(self) -> dict:
        """
        Arguments for the function
        """
        if self._args is None:
            return {}
        return self._args

    @property
    def extra_files(self) -> dict:
        """
        Returns the extra files set for this runner
        """
        return self._extra_files

    def read_local_files(self):
        self._logger.runtime("reading local runner files")

        def check_file_mtime(path: str) -> int:
            """
            Checks if a file at `path` exists, and is part of the current run

            Args:
                path:
                    filepath to check
            Returns:
                (int) timestamp if valid, -1 otherwise
            """
            self._logger.runtime(f"checking file {path}")
            # check if the file exists
            if not os.path.isfile(path):
                self._logger.runtime("file not found")
                return -1
            # check if the file is from after the most recent submission
            timestamp = int(os.path.getmtime(path))
            self._logger.runtime(
                f"checking mtime {timestamp} vs " f"submit time {self.last_submitted}"
            )
            if timestamp < self.last_submitted:
                self._logger.runtime("file is outdated")
                return -1

            return timestamp

        finished = False
        # first try read results
        timestamp = check_file_mtime(self.resultfile.local)
        if timestamp > 0:
            mtime = datetime.fromtimestamp(timestamp)
            self.insert_history(mtime, "resultfile created remotely")

            self.result = self.parent.serialiser.load(self.resultfile.local)
            finished = True

        # do the same for the error
        timestamp = check_file_mtime(self.errorfile.local)
        if timestamp > 0:
            mtime = datetime.fromtimestamp(timestamp)
            self.insert_history(mtime, "errorfile created remotely")

            with open(self.errorfile.local, "r") as o:
                error = o.read().strip()

            if error != "":
                self._logger.runtime("valid error found, storing")
                self.error = error.split("\n")[-1]

            finished = True

        return finished

    @property
    def result(self):
        """
        Result (If available)
        """
        if os.path.isfile(self.resultfile.local):
            success = self.read_local_files()

            if success:
                return self._result

        if hasattr(self, "_result"):
            try:
                if SERIALISED_STORAGE_KEY in self._result:
                    self._result = self.parent.serialiser.loads(self._result[1])
            except TypeError:
                pass

            return self._result
        return None

    @result.setter
    def result(self, result) -> None:
        """
        Creates and sets the result property, setting the state to "completed"

        Args:
            result:
                run result
        """
        self._result = result
        self.state = "completed"

    @property
    def error(self):
        """
        Error (If one exists)
        """
        if hasattr(self, "_error"):
            return self._error
        return None

    @error.setter
    def error(self, error) -> None:
        """
        Creates and sets the error property

        Args:
            error:
                run error string
        """
        self._error = error
        self.state = "completed"

    def clear_result(self) -> None:
        """
        Removes any results, and sets the state to "result wiped"
        """
        if hasattr(self, "_result"):
            del self._result
            self._logger.info("deleted _result property")
        if hasattr(self, "_error"):
            del self._error
            self._logger.info("deleted _error property")

        try:
            os.remove(self.resultfile.local)
            self._logger.info("removed result file")
        except FileNotFoundError:
            pass
        try:
            os.remove(self.errorfile.local)
            self._logger.info("removed error file")
        except FileNotFoundError:
            pass

        def remove_file(path):
            self._logger.info(f"attempting to clear result file {path}")
            try:
                os.remove(path)
                self._logger.info("Done")
            except FileNotFoundError:
                self._logger.info("file not found")

        remove_file(self.resultfile.local)
        remove_file(self.resultfile.remote)

        self.state = "result wiped"

    @property
    def state(self) -> str:
        """
        Returns the most recent runner state
        """
        return self._state

    @state.setter
    def state(self, newstate: str) -> None:
        """
        Update the state and store within the runner history
        """

        t = int(time.time())
        state_time = datetime.fromtimestamp(t)

        self.insert_history(state_time, newstate)
        self._state_time = t

        if newstate == Runner._submit_status["final"]:
            self._last_submitted = t
        self._state = newstate

    @property
    def last_updated(self) -> int:
        """
        Time that this runner state last changed
        """
        return self._state_time

    @property
    def last_submitted(self) -> int:
        """
        Time that this runnerwas last submitted
        """
        return self._last_submitted

    def format_time(self, t: datetime.time) -> str:
        """
        Format the datetime object into a dict key

        Args:
            t (datetime.time):
                time object to be formatted to string

        Returns:
            (str):
                formatted time
        """
        return t.strftime(_time_format)

    @property
    def history(self) -> dict:
        """
        State history of this runner
        """
        return self._history

    @property
    def status_list(self) -> list:
        """
        Returns a list of status updates
        """
        return list(self._history.values())

    def insert_history(self, t: datetime, newstate: str) -> None:
        """
        Insert a state into this runner's history

        Args:
            t (datetime.time):
                time this state change occurred
            newstate (str):
                status to update
        """
        if not isinstance(t, datetime):
            raise ValueError(f"time of type {type(t)} should be a datetime instance")

        base_timekey = self.format_time(t)
        idx = 0
        timekey = f"{base_timekey}/{idx}"
        while timekey in self._history:
            idx += 1

            timekey = f"{base_timekey}/{idx}"

        self._logger.info(
            f"({timekey}) updating runner {self.short_uuid} state -> {newstate}"
        )
        self._history[timekey] = newstate

    def run(self, dry_run: bool = False, **kwargs) -> None:
        """
        Perform a manual run

        .. warning::
            This method should be used sparingly, as it creates a Dataset
            object within the function from the Database. This is a costly
            process and potentially unstable.

        Args:
            dry_run (bool):
                create files, but do not run
        """

        parent = self.unserialise(self.database.find(self.parent.uuid))

        if not self._assess_run(**self._run_options):
            return None

        if os.name == "nt" and parent.url.is_local:
            raise RuntimeError(localwinerror)

        self._manual_run = True  # set internal flag for a manual run
        self.state = Runner._submit_status["initial"]
        self._write_runfile(parent)

        parent.transport.queue_for_push(
            self.runfile.name, self.local_dir, self.remote_dir
        )

        script = parent._script_sub(**self.run_args)

        self._write_script(parent.url.python, script)

        parent.transport.queue_for_push(
            self.jobscript.name, self.jobscript.local_dir, self.jobscript.remote_dir
        )

        cmd = (
            f"cd {self.jobscript.remote_dir} &&"
            f" {parent.url.submitter} {self.jobscript.name}"
        )

        if not dry_run:
            parent.transport.transfer()
            parent.url.cmd(cmd, asynchronous=False)
            self.state = Runner._submit_status["final"]
        else:
            parent.transport.wipe_transfers()
            self._logger.important(f"launch command: {cmd}")
            self.state = "dry run"

    def _assess_run(self, **kwargs) -> bool:

        send_extra = kwargs.pop("extra_files_send", [])
        self._extra_files["send"] += ensure_list(send_extra)

        recv_extra = kwargs.pop("extra_files_recv", [])
        self._extra_files["recv"] += ensure_list(recv_extra)

        self._run_options.update(kwargs)

        self._logger.important(f"assessing run for runner {self}", end="... ")
        self._logger.info("\nrun args:")
        self._logger.info(format_iterable(self.run_args))
        self._logger.info(f"current state is {self.state}")

        if self.run_option("force", False):
            self._logger.warning("force running")
            return True

        if self.state == Runner._submit_status["final"]:
            self._logger.warning("skipping already submitted run")
            return False

        # old skip test unpacks a runner from the database
        if self.is_finished and self.run_option("skip", True):
            self._logger.warning("skipping already completed run")
            return False

        self._logger.important("checks passed, running")
        return True

    def _write_runfile(self, dataset, write_file: bool = True) -> str:
        """
        Writes the python file which actually runs the function

        Args:
            dataset:
                parent dataset
            write_file (bool):
                skip writing runfile if false, used only for debugging purposes

        Returns:
            (str): runfile contents
        """

        self._logger.info(f"pre-running function {self.uuid}")

        if self._manual_run:
            self._logger.info("forcing dataset repo write")
            dataset._write_to_repo()

        # check if we have replaced args with a file, and use that if so
        if self._args_replaced:
            argstore = list(self.args.keys())[0]
            argline = f'kwargs = repo.load("' f'{self._relative_remote_dir(argstore)}")'
        else:
            argline = f"kwargs = {self.args}"

        runscript = [
            f"repo = __import__('{dataset.repofile}')",
            argline,
            f"result = repo.{dataset.function.name}(**kwargs)",
            f"repo.dump(result, '{self.resultfile.name}')",
        ]

        if dataset.is_child:
            runscript.insert(2, self._dependency_info["parent_import"])
        if dataset.is_parent:
            for line in self._dependency_info["child_submit"]:
                runscript.append(line)

        output = "\n".join(runscript)
        if write_file:
            if not os.path.isdir(self.local_dir):
                self._logger.info(f"creating local dir {self.local_dir}")
                os.makedirs(self.local_dir)
            with open(self.runfile.local, "w+") as o:
                o.write(output)
            self.state = "runfile written"
        else:
            self.state = "runfile write skipped"

        return output

    def _write_script(self, python: str, script, write_file: bool = True) -> str:
        """
        Writes the jobscript for this runner

        Args:
            python (str):
                python command to launch runfile
            script (str, URL.script):
                script header
        """
        tmp = []
        if isinstance(script, str):
            self._logger.debug("initialising script array with string input:")
            self._logger.debug(script)
            tmp.append(script)
        else:
            self._logger.info("script passed to runner requires substitution")
            tmp.append(script(**self.run_args))

        if self.run_dir and self.run_dir != self.remote_dir:
            self._logger.debug("run dir is separate to remote dir, appending extras")
            cmd = (
                f"pydir=$PWD\n"
                f"mkdir -p {self.run_dir} && "
                f"cd {self.run_dir} && "
                f"{python} ${{pydir}}/{self.runfile.name} 2>> {self.errorfile.name}"
            )
        else:
            runline = f"{python} {self.runfile.name} 2>> {self.errorfile.name}"
            self._logger.debug(f"directly appending line {runline}")
            cmd = runline  # run file

        tmp.append(cmd)

        path = self.jobscript.local
        output = "\n".join(tmp)
        if write_file:
            self._logger.info(f"writing run script to {path}")
            with open(path, "w+") as o:
                o.write(output)
                # Make sure the script ends with a line break
                o.write("\n")
            self.state = "jobscript written"
        else:
            self.state = "jobscript write skipped"

        return output

    @property
    def is_finished(self) -> bool:
        """
        Attempts to determine if this runner has completed its run

        Returns (bool):
            completion status
        """
        fin = hasattr(self, "_result") or hasattr(self, "_error")
        self._logger.info(
            f"checking finished state of runner {self.short_uuid} -> {fin}"
        )
        return fin

    def update_run_options(self, run_args: dict) -> None:
        """
        Update run args with dict `run_args`
        Args:
            run_args (dict):
                new run arguments

        Returns:
            None
        """
        self._logger.info("updating run options with new run args:")
        self._logger.info(format_iterable(run_args))

        self._run_options.update(run_args)

    def run_option(self, option: str, default=None):
        """
        Return a run option

        Args:
            option (str):
                key to search for
            default:
                default argument to provide to get

        Returns:
            option if available, else None
        """
        ret = self._run_options.get(option, default)
        self._logger.debug(f"getting run option {option}: {ret}")
        return ret

    @property
    def run_args(self) -> dict:
        """
        Display the run arguments

        Returns:
            (dict) run_args
        """
        return self._run_options



from remotemanager.logging import LoggingMixin
from remotemanager.storage import SendableMixin

import os


class TrackedFile(LoggingMixin, SendableMixin):

    __slots__ = ("_remote_path", "_local_path", "_file")

    def __init__(self, local_path, remote_path, file):

        self._remote_path = remote_path
        self._local_path = local_path
        self._file = file

    def __repr__(self):
        return self.local

    def __fspath__(self):
        return self.name

    @property
    def name(self):
        return self._file

    @property
    def remote(self):
        return os.path.join(self._remote_path, self.name)

    @property
    def local(self):
        return os.path.join(self._local_path, self.name)

    @property
    def remote_dir(self):
        return self._remote_path

    @property
    def local_dir(self):
        return self._local_path

    @property
    def content(self):
        with open(self.local, "r") as o:
            return o.read()



import functools
import logging
import operator
import warnings

import yaml

from remotemanager.utils import get_version, ensure_filetype, recursive_dict_update
from remotemanager.utils.version import Version


class Database:
    """
    Database file handler for use in the Dataset.

    .. warning::
        Interacting with this object directly could cause unstable behaviour.
        It is best to allow Dataset to handle the Database

    Args:
        file (str):
            filename to write to
    """

    _versionkey = "~database-version~"

    def __init__(self, file):

        file = ensure_filetype(file, "yaml")

        self._logger = logging.getLogger(__name__ + ".Database")

        self._path = file
        self._storage = self._read()
        self._tree = None

    def _read(self) -> dict:
        """
        Wipes the tree cache and reads from the database file, creating it if
        it does not already exist.
        """
        self._tree = None  # wipe tree cache
        self._logger.info(f"reading {self._path}")
        try:
            with open(self._path, "r") as o:
                data = yaml.safe_load(o)
        except FileNotFoundError:
            self._logger.warning("file not found, creating")
            data = {Database._versionkey: get_version()}
            self._write(data)

        try:
            version = data.pop(Database._versionkey)
        except KeyError:
            version = "0.0.0"

        self._logger.info(f"database file version: {version}")

        if Version(version) < get_version():
            msg = (
                f"database file (version {version}) is less than "
                f"current remotemanager install version"
            )
            warnings.warn(msg)
            self._logger.warning(msg)

        return data

    def _write(self, data: dict):
        """
        Clears the tree cache and writes to the file

        Args:
            data (dict):
                writes `data` to the database file
        """
        self._tree = None  # wipe tree cache

        data[Database._versionkey] = get_version()

        with open(self._path, "w+") as o:
            yaml.dump(data, o)

    def read(self):
        """
        Read the database file
        """
        self._storage = self._read()

    def write(self):
        """
        Write the database to file
        """
        self._write(self._storage)

    def update(self, payload):
        """
        Update the database with payload (dict)

        Args:
            payload (dict):
                Dictionary to recursively update with. Usually called with the
                output of object.pack()
        """

        self._logger.info("updating stored info")

        self._storage = recursive_dict_update(self._storage, payload)
        self.write()

    @property
    def path(self):
        """
        Path to current database file
        """
        return self._path

    @property
    def tree(self) -> list:
        """
        Returns a list of path-like strings for the stored database dict

        Returns (list):
            List of path-like strings
        """
        if self._tree is not None:
            return self._tree
        self._tree = self.climb(self._storage)
        return self._tree

    def climb(self, data: dict, branch: list = None) -> list:
        """
        "climb" a dictionary tree, returning a list of path-like strings for
        each element

        .. note::
            This method is intended for use within the `tree` property, and
            could cause unintended behaviour if called directly, though will
            allow you to produce a `tree` like list for any dictionary. Use
            with caution.

        Args:
            data (dict):
                dictionary to treat
            branch (list):
                current branch, used for recursion

        Returns (list):
            list of path-like strings
        """

        if branch is None:
            branch = []

        joinchar = "/"

        ret = []

        if len(data) == 0:
            ret.append(joinchar.join(branch))

        for key in data:
            tmp = []
            if isinstance(data[key], dict):
                tmp.append(key)
                ret += self.climb(data[key], branch + tmp)
            else:
                ret.append(joinchar.join(branch + [key]))
        return ret

    def find(self, key: str):
        """
        Find the first instance of key within the database tree

        Args:
            key (str):
                key (uuid) to look for

        Returns:
            database tree below key
        """

        def get_minimum_path(p, k):
            fullpath = p.split("/")
            return fullpath[: fullpath.index(k) + 1]

        target = []
        for path in self.tree:
            if key in path:
                target = get_minimum_path(path, key)
                break
        data = chain_get(self._storage, target)

        return data


def chain_get(d: dict, keys: str):
    """
    Get item from a nested dict using a list of keys

    Args:
        d (dict):
            nested dict to query
        keys (list):
            list of keys to use

    Returns:
        item from nested dict `d` at [list, of, keys]
    """
    return functools.reduce(operator.getitem, keys, d)



import gc
import os
import math
import time

# need __version__, is there a better way than importing the whole package?
import remotemanager

_time_format = "%Y-%m-%d %H:%M:%S"


def ensure_list(inp=None) -> list:
    """
    Ensure that `inp` is list-type

    Args:
        inp:
            list, string, object to be processed

    Returns (list):
        Either inserts the object into a list, or returns the list-like inp
    """
    if inp is None:
        return []
    elif isinstance(inp, (list, tuple, set)):
        return list(inp)
    elif isinstance(inp, str):
        return [inp]
    elif inp.__class__.__name__ == "TrackedFile":
        return [inp]
    return list(inp)


def ensure_filetype(file, target_type):
    """
    Ensure that `file` is of type `type`

    Args:
        file (str):
            filename
        target_type (str):
            filetype to enforce

    Returns (str):
        filename of type `type`
    """
    fname, ftype = os.path.splitext(file)

    target_type = target_type.strip(".")

    return f"{fname}.{target_type}"


def ensure_dir(dir):
    """
    Ensure that string path to `dir` is correctly formatted
    ONLY ensures that the folder name ends with a "/", does not produce an
    abspath

    Args:
        dir (str):
            path to dir

    Returns (str):
        ensured dir path
    """
    if not isinstance(dir, str):
        dir = str(dir)

    return os.path.join(dir, " ").strip()


def safe_divide(a: [float, int], b: [float, int]) -> [float, int]:
    """
    Always-fit division. Rounds up after division, returns >= 1

    Args:
        a (num):
            numerator
        b (num):
            denominator

    Returns:
        result of a/b division, unless result returns <1, else 1
    """
    if b == 0:
        return 1
    r = math.ceil(a / b)
    return max(r, 1)


def get_version() -> str:
    """
    Gets the current package version from __init__.py
    """
    return remotemanager.__version__


def recursive_dict_update(d1: dict, d2: dict) -> dict:
    """
    Update d1 with all the keys of d2

    Args:
        d1 (dict):
            dictionary to be updated
        d2 (dict):
            dictionary to update with

    Returns (dict):
        updated d1
    """
    if not isinstance(d1, dict):
        return d2

    for k, v in d2.items():
        if isinstance(v, dict) and k in d1:
            d1[k] = recursive_dict_update(d1[k], d2[k])
        else:
            d1[k] = v

    return d1


def reverse_index(inplist: list, term):
    """
    return index of last occurrence of `term` in `inplist`

    Args:
        inplist (list):
            list to index
        term:
            object to index

    Returns (int):
        forward index of item
    """
    return len(inplist) - inplist[::-1].index(term) - 1


def integer_time_wait(offset: float = 0.0) -> None:
    """
    wait for an integer unix time + offset

    This function exists to increase the reproducibility of the tests

    Args:
        offset (float):
            extra offset to wait for
    """

    t0 = time.time()

    wait = 1 - math.fmod(t0, 1)

    time.sleep(wait + offset)

    return None


def object_from_uuid(uuid: str, name: str):
    """
    Search the memory heap for an object of uuid `uuid` and classname
    `name`

    for example;

    >>> recovered_dataset = object_from_uuid('<uuid>', 'Dataset')

    Args:
        uuid:
            uuid target
        name:
            classname to search for

    Returns:
        object instance matching input, if it exists
    """

    for obj in gc.get_objects():
        try:
            if obj.__class__.__name__ == name:
                if obj.uuid == uuid:
                    return obj
        except AttributeError:
            pass

    raise ValueError(f"could not find {object} in memory with UUID {uuid}")



import dill

from remotemanager.serialisation.serial import serial


class serialdill(serial):
    """
    subclass of serial, implementing dill methods
    """

    def dumps(self, obj):
        return dill.dumps(obj)

    def loads(self, string):
        return dill.loads(string)

    @property
    def callstring(self):
        return "dill"

    @property
    def bytes(self):
        return True



import yaml

from remotemanager import URL

from remotemanager.connection.computers.options import optional, required
from remotemanager.storage.function import Function


def format_time(time):
    if not isinstance(time, int):
        return time
    mins = time // 60
    hours = mins // 60

    secstring = str(time % 60).rjust(2, "0")
    minstring = str(mins % 60).rjust(2, "0")
    hourstring = str(hours).rjust(2, "0")

    return f"{hourstring}:{minstring}:{secstring}"


class BaseComputer(URL):
    """
    Base computer module for HPC connection management.

    Extend this class for connecting to your machine
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.shebang = "#!/bin/bash"
        self._parser = None

        self.choice_groupings = []

        self._extra = ""
        self._internal_extra = ""

    def __setattr__(self, key, value):
        """
        If the set `key` attribute is an MPI option, instead set the `value`
        of that attribute

        Args:
            key:
                attribute name to set
            value:
                value to set to

        Returns:
            None
        """
        if key in self.__dict__:
            if isinstance(getattr(self, key), (optional, required)):
                getattr(self, key).value = value
                return

        object.__setattr__(self, key, value)

    def pack(self, *args, **kwargs):
        if not isinstance(self._parser, (Function, type(None))):
            self._parser = Function(self._parser)

        return super().pack(*args, **kwargs)

    @classmethod
    def from_dict(cls, spec: dict, **url_args):
        """
        Create a Computer class from a `spec` dictionary. The required values are:

        - resources:
            a dict of required resources for the machine (mpi, nodes, queue, etc)
        - resource_parser:
            a function which takes a dictionary of {resource: Option}, returning a list
            of valid jobscript lines

        You can also provide some optional arguments:

        - required_or:
            list of resources, ONE of which is required. Note that this must be a
            _list_ of dicts, one for each or "block"
        - optional_resources:
            as with `resources`, but these will not be stored as required values
        - optional_defaults:
            provide defaults for the names given in optional_resources. When adding the
            optional arg, the optional_defaults will be checked to see if a default is
            provided
        - host:
            machine hostname. Note that this _must_ be available for a job run, but if
            not provided within the spec, can be added later with url.host = 'hostname'
        - submitter:
            override the default submitter
        - python:
            override the default python
        - extra:
            any extra lines that should be appended after the resource specification.
            Note that this includes module loads/swaps, but this can be specified on a
            per job basis, rather than locking it into the `Computer`

        The `resources` specification is in `notebook`:`machine` order. That is to say
        that the `key` is what will be required in the _notebook_, and the `value` is
        what is placed in the jobscript:

        >>> spec = {'resources': {'mpi': 'ntasks'}, ...}
        >>> url = BaseComputer.from_dict(spec)
        >>> url.mpi = 12
        >>> url.script()
        >>> "--ntasks=12"

        Args:
            spec (dict):
                input dictionary
            url_args:
                any arguments to be passed directly to the created `url`

        Returns:
            Computer class as per input spec
        """

        def set_resource_spec(field, flag, opt=False, default=None) -> None:
            """
            Set the resource requester in the computer

            Args:
                field:
                    name under which to apply
                flag:
                    placeholder arguments
                opt:
                    True if optional argument. Required as default can be None
                default:
                    default arg (if any)
            """
            if isinstance(flag, str):
                args = {"flag": flag}

            else:
                args = flag

            if opt:
                args["default"] = default
                attribute = optional(**args)
            elif "default" in args:
                attribute = optional(**args)
            else:
                attribute = required(**args)

            computer.__setattr__(field, attribute)

        # create a new class, primed with the hostname and parser
        computer = cls(host=spec.get("host", None), **url_args)
        computer.parser = spec.get("resource_parser")
        computer.submitter = spec.get("submitter", None)
        computer.python = spec.get("python", None)

        # add the resource storing objects to the class
        required_resources = spec.get("resources")
        selected_resources = spec.get("required_or", [])
        optional_resources = spec.get("optional_resources", {})

        for field, flag in required_resources.items():
            set_resource_spec(field, flag, opt=False)

        for group in selected_resources:
            allnames = list(group)
            computer.choice_groupings.append(allnames)

            for field, flag in group.items():
                set_resource_spec(field, flag, opt=False)

        defaults = spec.get("optional_defaults", {})
        for field, flag in optional_resources.items():
            default = None
            if field in defaults:
                default = defaults[field]
            set_resource_spec(field, flag, opt=True, default=default)

        # add any extra text
        computer.extra = spec.get("extra", "")

        return computer

    def to_dict(self) -> dict:
        """
        Generate a spec dict from this Computer

        Returns:
            dict
        """

        def create_entry(obj):
            data = {"flag": obj.flag}

            if obj.min is not None:
                data["min"] = obj.min
            if obj.max is not None:
                data["max"] = obj.max
            if getattr(obj, "default", None) is not None:
                data["default"] = obj.default

            return data

        spec = {}

        if self.host:
            spec["host"] = self.host

        spec["extra"] = getattr(self, "extra", "")
        spec["submitter"] = getattr(self, "submitter", "bash")

        required = {}
        required_or = []
        optional = {}
        defaults = {}

        flat = []
        for group in self.choice_groupings:
            tmp = {}
            for name in group:
                flat.append(name)
                obj = getattr(self, name)
                tmp[name] = create_entry(obj)

            required_or.append(tmp)

        for name in self.arguments:
            if name in flat:
                continue
            obj = getattr(self, name)
            if name in self.required:
                required[name] = create_entry(obj)
            else:
                optional[name] = create_entry(obj)

                if obj.default:
                    defaults[name] = obj.default

        spec["resources"] = required

        if len(required_or) > 0:
            spec["required_or"] = required_or

        if len(optional) > 0:
            spec["optional_resources"] = optional

        if len(defaults) > 0:
            spec["optional_defaults"] = defaults

        # avoids a strange error where the parser source cant be found.
        # likely loads it into memory where `inspect` can access it
        self.parser
        if isinstance(self._parser, Function):
            spec["resource_parser"] = self._parser
        else:
            spec["resource_parser"] = Function(self.parser)

        return spec

    @classmethod
    def from_yaml(cls, filepath: str, **url_args):
        if isinstance(filepath, str):
            with open(filepath, "r") as o:
                data = yaml.safe_load(o)
        else:
            data = yaml.safe_load(filepath)

        # convert the parser back into a function
        data["resource_parser"] = Function.unpack(data["resource_parser"])

        return cls.from_dict(data, **url_args)

    def to_yaml(self, filepath):
        data = self.serialise(self.to_dict())

        if isinstance(filepath, str):
            with open(filepath, "w+") as o:
                yaml.dump(data, o)
        else:
            yaml.dump(data, filepath)

    @classmethod
    def from_repo(
        cls,
        name: str,
        branch: str = "main",
        repo: str = "https://gitlab.com/l_sim/remotemanager-computers/",
        **url_args,
    ):
        """
        Attempt to access the remote-computers repo, and pull the computer with name
        `name`

        Args:
            name (str):
                computer name to target
            branch (str):
                repo branch (defaults to main)
            repo (str):
                repo web address (defaults to main l_sim repo)

        Returns:
            BaseComputer instance
        """
        import requests
        from remotemanager.utils import ensure_filetype

        def download_file(file_url, filename):
            response = requests.get(file_url)

            if response.status_code == requests.codes.ok:
                # Save the file
                with open(filename, "wb") as file:
                    file.write(response.content)
                print(f"Grabbed file '{filename}'")
            else:
                raise RuntimeError(f"Could not find a file at: {file_url}")

        filename = ensure_filetype(name, "yaml").lower()
        url = f"{repo}-/raw/{branch}/storage/{filename}"

        print(f"polling url {url}")

        download_file(url, filename)

        return cls.from_yaml(filename, **url_args)

    @property
    def parser(self):
        if self._parser is None:
            return None
        if not isinstance(self._parser, Function):
            self._parser = Function(self._parser)
        return self._parser.object

    @parser.setter
    def parser(self, parser):
        if not isinstance(parser, Function):
            parser = Function(parser)
        self._parser = parser

    @property
    def arguments(self):
        return [
            k for k, v in self.__dict__.items() if isinstance(v, (optional, required))
        ]

    @property
    def argument_dict(self):
        return {k.strip(): getattr(self, k) for k in self.arguments}

    @property
    def required(self):
        """

        Returns the required arguments
        """
        return [k for k, v in self.__dict__.items() if isinstance(v, required)]

    @property
    def missing(self):
        """
        Returns the currently missing arguments
        """
        missing = [k for k in self.required if not getattr(self, k)]

        # check the groupings to make sure we're note raising any false positives
        for group in self.choice_groupings:
            if any(name not in missing for name in group):
                # at least one of the names in the group is present in the computer,
                # so we're fine to continue
                for name in group:
                    try:
                        missing.pop(missing.index(name))
                    except ValueError:
                        pass

        return missing

    @property
    def valid(self):
        return len(self.missing) == 0

    def update_resources(self, **kwargs):
        """
        Set any arguments passed to the script call

        Args:
            **kwargs:
                kwarg updates e.g. mpi=128
        Returns:
            None
        """
        for k, v in kwargs.items():
            setattr(self, k, v)

    def resources_block(self, **kwargs):

        self.update_resources(**kwargs)

        if not self.valid:
            raise RuntimeError(f"missing required arguments: {self.missing}")

        if self._parser is None:
            return []

        return self.parser(self.argument_dict)

    @property
    def extra(self):
        return self._internal_extra + self._extra

    @extra.setter
    def extra(self, external):
        self._extra = external

    def script(self, **kwargs) -> str:
        """
        Takes job arguments and produces a valid jobscript

        Returns:
            (str):
                script
        """
        script = [self.shebang]

        script += self.resources_block(**kwargs)

        if self.extra is not None:
            script += [self.extra]

        return "\n".join(script)



class Version:
    """
    Lightweight temporary class for version comparison.

    Create at least one instance of Version to compare two semantic versions:

    >>> v = Version('1.5.2')
    >>> v < '4.1.0'
    >>> True

    Args:
        ver (str):
            Semantic version in x.y.z form
    """

    def __init__(self, ver):
        self._version = ver

        self._major, self._minor, self._patch = [int(v) for v in ver.split(".")]

    def _coerce_version(self, other):
        if not isinstance(other, Version):
            other = Version(other)
        return other

    def _compare(self, other):
        if self.major != other.major:
            return -1 if self.major < other.major else 1
        if self.minor != other.minor:
            return -1 if self.minor < other.minor else 1
        if self.patch != other.patch:
            return -1 if self.patch < other.patch else 1
        return 0

    def __eq__(self, other):
        other = self._coerce_version(other)
        return self._compare(other) == 0

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        other = self._coerce_version(other)
        return self._compare(other) < 0

    def __gt__(self, other):
        other = self._coerce_version(other)
        return self._compare(other) > 0

    def __le__(self, other):
        other = self._coerce_version(other)
        return self._compare(other) <= 0

    def __ge__(self, other):
        other = self._coerce_version(other)
        return self._compare(other) >= 0

    def __repr__(self):
        return f"{self.major}.{self.minor}.{self.patch}"

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor

    @property
    def patch(self):
        return self._patch

    @property
    def version(self):
        return self._version



from enum import Enum

from remotemanager.storage.sendablemixin import SendableMixin


class Verbosity(SendableMixin):
    """
    Class to store verbosity information

    Initialise with Verbosity(level), where level is the integer level

    args:
        level (int, str):
            level above which to print
    """

    def __init__(self, level):
        if level is None:
            level = 4

        # see if the level passed is already a Verbose instance
        if isinstance(level, self.__class__):
            level = level.value

        # first try to get level via from an int, falling back on string
        try:
            level = int(level)
        except ValueError:
            try:
                level = getattr(_levels, level.lower()).value
            except AttributeError:
                # raise _levels missing error with invalid level
                _levels(level)

        self._value = level
        self._name = _levels(level).name

    def __repr__(self):
        return self.name

    def __bool__(self):
        return self.value != 0

    @property
    def value(self):
        return self._value

    @property
    def level(self):
        return self._value

    @property
    def name(self):
        return self._name

    def print(self, message: str, atlevel: int = 2, end: str = "\n"):
        """
        Request that a message be printed. Compares against the set
        verbosity level before printing.

        Args:
            message (str):
                message to print
            atlevel (int):
                If this number is higher priority than (lower numeric value)
                (or equal to) the set limit, print
            end (str):
                print(..., end= ...) hook
        """
        # print(f'request {message[:24]} @ {atlevel}')
        if self and self.level <= atlevel:
            print(message, end=end)


class _levels(Enum):
    none = 0
    debug = 1
    runtime = 2
    info = 3
    important = 4
    warning = 5
    error = 6
    critical = 7

    @classmethod
    def _missing_(cls, value):
        msg = [f"{value} is not a valid Verbosity:"]

        for v in _levels:
            msg.append(f"\t{v.value}: {v.name}")

        raise ValueError("\n".join(msg))



"""
Handles file transfer via the `rsync` protocol
"""
import pathlib

from remotemanager.transport.transport import Transport


class rsync(Transport):
    """
    Class for `rsync` protocol

    Args:
        checksum (bool):
            Adds checksum arg, which if True will add --checksum flag to
            parameters
    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # flags can be exposed, to utilise their flexibility
        flags = kwargs.pop("flags", "auv")
        self.flags = flags

        if kwargs.get("checksum", False):
            print("adding checksum to rsync")
            self.flags += "--checksum"

        self._logger.info("created new rsync transport")

        self._cmd = "rsync {flags} {password}{inner_dir}{primary} {secondary}"

    def cmd(self, primary, secondary):

        if self.url.passfile and self.url.keyfile:
            raise RuntimeError(
                "rsync appears to have an issue when "
                "specifying sshpass AND ssh-key. Either set up "
                "your ssh config and remove the keyfile or use "
                "transport.scp"
            )

        def get_remote_dir(path):
            if ":" not in path:
                return path
            return path.split(":")[1]

        password = ""
        if self.url.passfile is not None:
            password = f'--rsh="sshpass -f {self.url.passfile} ssh" '

        cmd = self._cmd
        inner_dir = ""
        if len(pathlib.Path(secondary).parts) > 1:
            # the target is a nested dir. If the whole tree doesn't exist,
            # rsync will throw an error
            if ":" in secondary:
                # target is a remote folder, use the --rsync-path hack
                inner_dir = (
                    f'--rsync-path="mkdir -p ' f'{get_remote_dir(secondary)} && rsync" '
                )
            else:
                cmd = f"mkdir -p {secondary} && {self._cmd}"

        base = cmd.format(
            flags=self.flags,
            password=password,
            primary=primary,
            secondary=secondary,
            inner_dir=inner_dir,
        )
        self._logger.debug(f'returning formatted cmd: "{base}"')
        return base
